unit IEPackages;
//=== File Prolog ============================================================
//	This code was developed by RiverSoftAVG.
//
//--- Notes ------------------------------------------------------------------
//
//--- Development History  ---------------------------------------------------
//
//      01/2001 T. Grubb     Modified TIEInnerFunction to be more generic.  Now, you can use
//                           TIEInnerFunction to connect to ANY TObject.
//                           Also, one component
//                           can now act as multiple IUserFunctions using one TIEInnerFunction object
//                           per IUserFunction connection
//                               - Changed Constructor to accept TComponent instead of TCustomUserFunction
//                               - Changed Owner property to TComponent from TCustomUserFunction\
//                               - Added OnCall event for non-TCustomUserFunction components
//                               - Changed Call method to call TCustomUserFunction.Call method directly
//                                 if Owner is TCustomUserFunction.  Otherwise, the new OnCall event
//                                 is triggered
//                               - Fixed TIEInnerFunction Notification method to check properly for
//                                 removal of owner (Someone could have overridden TIEInnerFunction
//                                 and connected to the TIEFunctionNexus causing this method to
//                                 erroneously nil the Owner)
//      11/2000 T. Grubb     Moved TUserPackage.CreateFunctions from constructor
//                           to AfterConstruction for improved C++Builder compatibility
//	04/2000	T. Grubb     Initial version.
//
//      File Contents:
//           TBaseReasonerList        Interface list that only works with IReasoners
//                                    and notifies the owner when things change
//           TCustomReasonerOwner     Component that owns a list of reasoners,
//                                    base class for TUserPackage
//           TCustomUserFunction      Component for an expert system (ES) run-time function
//                                    available at design time
//           TFunctionCall            Object that encapsulates a call to a function
//           TIEFunctionNexus         Glue component between TCustomUserFunction and
//                                    TIEInnerFunction
//           TIEInnerFunction         TIEUserFunction descendant intended to be
//                                    owned by another object (notifies it when
//                                    function is called
//           TIEUserFunction          Base class for all IUserFunctions in the
//                                    inference engine architecture
//           TUserFunctionList        Functions list for a TUserPackage
//           TReasonerList            Reasoners list for a TCustomReasonerOwner (TIList descendant)
//           TUserPackage             a package or collection of functions
//           TInterpretedFunction     TIEUserFunction descendant for a function that
//                                    is implemented by other functions and is
//                                    interpreted at ES run-time
//           TUserFunction            TCustomUserFunction descendant that exposes properties
//           TUserFunctionList        List of user functions, tracks them for its
//                                    TUserPackage
//
//
//--- Warning ----------------------------------------------------------------
//	This software is property of RiverSoftAVG. Unauthorized use or
//      duplication of this software is strictly prohibited. Authorized users
//      are subject to the following restrictions:
//	*	RiverSoftAVG is not responsible for
//		any consequence of the use of this software.
//	*	The origin of this software must not be misrepresented either by
//		explicit claim or by omission.
//	*	Altered versions of this software must be plainly marked as such.
//	*	This notice may not be removed or altered.
//
//      (c) 2000-2001, Thomas G. Grubb
//
//=== End File Prolog ========================================================

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  IEInterfaces, InferenceEngine, InterfaceCollections, Structures, IEConsts,
  CommonInterfaces, IEContext{$IFDEF VER130}, Contnrs {$ELSE}, GLists {$ENDIF};

{$ObjExportAll On}

type

  { Notes:
          Packages are other components that can be assigned to the inference engine
          at design time.  Packages are libraries of functions that can be used to
          expand the Inference Engine syntax.

          Fact Sets are CLIPS DefFacts, which are collections of facts to be instantiated
          when the engine is reset.

  }

  TCallEvent = procedure ( Sender: TObject; FunCall: IFunCall;
                           Context: TIEContext; var Result: IIEValue ) of object;

  TCustomReasonerOwner = class;
  TIEUserFunction = class;
  TBaseReasonerList = class(TIList)
  { Purpose: To create an interface list that only works with IReasoners
    and notifies the owner when things change }
  private
    { Private Declarations }
    FOwner: TIEUserFunction;
    function GetItem(Index: Integer): IReasoner;
    procedure SetItem(Index: Integer; const Value: IReasoner);
  protected
    { Protected Declarations }
    procedure Notify(Item: IUnknown; Action: TListNotification); override;
    property Owner: TIEUserFunction read FOwner;
  public
    { Public Declarations }
    constructor Create( AOwner: TIEUserFunction ); virtual;
    function Add(Reasoner: IReasoner): Integer;
    function Remove(Reasoner: IReasoner): Integer;
    function IndexOf(Reasoner: IReasoner): Integer;
    procedure Insert(Index: Integer; Reasoner: IReasoner);
    property Items[Index: Integer]: IReasoner read GetItem write SetItem; default;
  published
    { Published Declarations }
  end; { TBaseReasonerList }

  TReasonerList = class(TBaseReasonerList)
  { Purpose: To create an interface list that only works with IReasoners
    and notifies the owner when things change }
  private
    { Private Declarations }
    FOwner: TCustomReasonerOwner;
  protected
    { Protected Declarations }
    procedure Notify(Item: IUnknown; Action: TListNotification); override;
    property Owner: TCustomReasonerOwner read FOwner;
  public
    { Public Declarations }
    constructor Create( AOwner: TCustomReasonerOwner ); reintroduce; virtual;
  published
    { Published Declarations }
  end; { TReasonerList }

  TIEUserFunction = class(TPersistent, IUserFunction, IDTStreamable, IUnknown)
  { Purpose: To implement the IUserFunction interface and serve as the base class
    for user functions in the TUserPackage component }
  private
    { Private Declarations }
    FRefCount: Integer;
    FMinArgumentCount: Integer;
    FMaxArgumentCount: Integer;
    FComment: String;
    FReasoners: TBaseReasonerList;
    FFunctionName: TIEName;
    procedure SetReasoners(const Value: TBaseReasonerList);
  protected
    { Protected Declarations }
    procedure ReasonerNotification( Item: IReasoner; Action: TListNotification ); virtual;
    procedure Call( FunCall: IFunCall; Context: TIEContext;
              var Result: IIEValue ); overload; virtual;
    procedure SetMaxArgumentCount( const Value: Integer ); virtual;
    procedure SetMinArgumentCount( const Value: Integer ); virtual;
    { IUnknown }
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    { IUserFunction }
    function AddReasoner( AReasoner: IReasoner ): Integer;
    procedure ClearReasoners;
    procedure DeleteReasoner( Index: Integer ); overload;
    procedure DeleteReasoner( AReasoner: IReasoner ); overload;
    function IndexOfReasoner( AReasoner: IReasoner ): Integer;
    function GetMaxArgumentCount: Integer; virtual;
    function GetMinArgumentCount: Integer; virtual;
    function GetReasoner(  Index: Integer  ): IReasoner;
    function GetReasonerCount: Integer;
    function GetComment: String;
    function GetFunctionName: TIEName;
    function GetToString( const Reasoner: IReasoner ): String;
    procedure SetComment( const Value: String );
    procedure SetFunctionName(const Value: TIEName); virtual;
    procedure Call( FunCall: IFunCall; Context: TObject; var Result: IIEValue ); overload;
    { IDTStreamable }
    function GetClassName: String;
    procedure ReadData(Reader: TReader); virtual;
    procedure WriteData(Writer: TWriter); virtual;
  public
    { Public Declarations }
    procedure Assign( Source: TPersistent ); override;
    constructor Create; overload; virtual;
    constructor Create( AName: TIEName ); overload;
    destructor Destroy; override;
    procedure AfterConstruction; override;
    procedure BeforeDestruction; override;
    class function NewInstance: TObject; override;
    property RefCount: Integer read FRefCount;
    property Reasoners: TBaseReasonerList read FReasoners write SetReasoners;
    property ReasonerCount: Integer read GetReasonerCount;
    property Reasoner[ Index: Integer ]: IReasoner read GetReasoner;
  published
    { Published Declarations }
    property MaxArgumentCount: Integer read FMaxArgumentCount;
    property MinArgumentCount: Integer read FMinArgumentCount;
    property Comment: String read FComment write SetComment;
    property FunctionName: TIEName read FFunctionName write SetFunctionName;
  end; { TIEUserFunction }
  TIEUserFunctionClass = class of TIEUserFunction;

  TIEFunctionNexus = class;
  TCustomUserFunction = class;
  TIEInnerFunction = class(TIEUserFunction)
  { Purpose: To implement a user function that is contained and synchronized with
    a component.  If Owner is a TCustomUserFunction, this object when CALLed will
    directly call the TCustomUserFunction.Call method.  Otherwise, it is up to
    the owner to assign a method to the OnCall event }
  private
    { Private Declarations }
    FOwner: TObject;
    FNexus: TIEFunctionNexus;
    FOnCall: TCallEvent;
  protected
    { Protected Declarations }
    procedure Call( FunCall: IFunCall; Context: TIEContext;
              var Result: IIEValue ); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); virtual;
    property Nexus: TIEFunctionNexus read FNexus;
  public
    { Public Declarations }
    constructor Create( AOwner: TObject ); reintroduce; virtual;
    destructor Destroy; override;
    property Owner: TObject read FOwner;
    property OnCall: TCallEvent read FOnCall write FOnCall;
  published
    { Published Declarations }
  end; { TIEInnerFunction }

  TIEFunctionNexus = class(TComponent)
  { Purpose: To implement an object that can listen for free component notifications
    for a non-component TIEInnerFunction object }
  private
    { Private Declarations }
    FInnerFunction: TIEInnerFunction;
  protected
    { Protected Declarations }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    property InnerFunction: TIEInnerFunction read FInnerFunction;
  public
    { Public Declarations }
    constructor Create( InnerFunction: TIEInnerFunction ); reintroduce;
  published
    { Published Declarations }
  end; { TIEFunctionNexus }

  TCustomReasonerOwner = class( TComponent )
  { Purpose:  To own a list of reasoners.  Base class for TUserPackage
  }
  private
    { Private Declarations }
    FReasoners: TReasonerList;
    procedure SetReasoners(const Value: TReasonerList);
  protected
    { Protected Declarations }
    procedure ReasonerNotification( Item: IReasoner; Action: TListNotification ); virtual;
    property Reasoners: TReasonerList read FReasoners write SetReasoners;
  public
    { Public Declarations }
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
  published
    { Published Declarations }
  end; { TCustomReasonerOwner }

  TCustomUserFunction = class(TComponent)
  { Purpose: Base class for users functions to be added to a reasoning engine at
    design time.  It uses the TIEInnerFunction class to actually implement the
    IUserFunction interface and talks with it through accessor properties.  It only
    manages an interface reference to the TIEInnerFunction class, this allows the
    inference engine to properly manage the interface references without worrying
    about the TCustomUserFunction component being freed by the component memory
    scheme, i.e., even if the TCustomUserFunction is freed, the inner function will
    not be freed until the inference engine is finished with it. }
  private
    { Private Declarations }
    FOnCall: TCallEvent;
    FEngine: TInferenceEngine;
    FInnerFunction: IUserFunction;             // Used for design time setting of the reasoner
    procedure SetEngine( const Value: TInferenceEngine );
    function GetComment: String;
    function GetFunctionName: TIEName;
    function GetMaxArgumentCount: Integer;
    function GetMinArgumentCount: Integer;
    procedure SetComment(const Value: String);
    procedure SetFunctionName(const NewName: TIEName);
    procedure SetMaxArgumentCount(const Value: Integer);
    procedure SetMinArgumentCount(const Value: Integer);
  protected
    { Protected Declarations }
    procedure Call( FunCall: IFunCall; Context: TIEContext;
              var Result: IIEValue ); dynamic;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    property Engine: TInferenceEngine read FEngine write SetEngine;
    property OnCall: TCallEvent read FOnCall write FOnCall;
  public
    { Public Declarations }
    procedure Assign( Source: TPersistent ); override;
    constructor Create( AOwner: TComponent ); overload; override;
    constructor Create( AOwner: TComponent; AName: TIEName ); reintroduce; overload;
    destructor Destroy; override;
    property InnerFunction: IUserFunction read FInnerFunction;
    property Comment: String read GetComment write SetComment;
    property FunctionName: TIEName read GetFunctionName write SetFunctionName;
    property MaxArgumentCount: Integer read GetMaxArgumentCount write SetMaxArgumentCount;
    property MinArgumentCount: Integer read GetMinArgumentCount write SetMinArgumentCount;
  published
    { Published Declarations }
  end; { TCustomUserFunction }
  TUserFunctionClass = class of TCustomUserFunction;

  TUserFunction = class(TCustomUserFunction)
  { Purpose: Base class for users functions to be added to a reasoning engine at
    design time.  It uses the TIEInnerFunction class to actually implement the
    IUserFunction interface and talks with it through accessor properties.  It only
    manages an interface reference to the TIEInnerFunction class, this allows the
    inference engine to properly manage the interface references without worrying
    about the TCustomUserFunction component being freed by the component memory
    scheme, i.e., even if the TCustomUserFunction is freed, the inner function will
    not be freed until the inference engine is finished with it. }
  private
    { Private Declarations }
  protected
    { Protected Declarations }
  public
    { Public Declarations }
  published
    { Published Declarations }
    property Comment;
    property FunctionName;
    property MaxArgumentCount;
    property MinArgumentCount;
    property Engine;
    property OnCall;
  end; { TUserFunction }

  TUserPackage = class;
  TFunctionInterfaceHeap = Array[0..99] of IUnknown;
  TUserFunctionList = class(TIList)
  { Purpose: To create a list that only works with IUserFunctions
    and notifies the package when things change.  It stores the functions as
    objects so that they may be streamed out and in but increments their reference
    counts so that they won't be freed without its knowledge }
  private
    { Private Declarations }
    FPackage: TUserPackage;
{$IFDEF DEMO}
    FHeap: TFunctionInterfaceHeap;
{$ENDIF}
    function GetItem(Index: Integer): IUserFunction;
    procedure SetItem(Index: Integer; const Value: IUserFunction);
  protected
    { Protected Declarations }
    function GetToString: String; virtual;
    procedure Notify(Item: IUnknown; Action: TListNotification); override;
  public
    { Public Declarations }
    constructor Create( Package: TUserPackage ); reintroduce; overload;
    function Add(AFunction: IUserFunction): Integer;
    function Remove(AFunction: IUserFunction): Integer;
    function IndexOf(AFunction: IUserFunction): Integer;
    procedure Insert(Index: Integer; AFunction: IUserFunction);
    property Items[Index: Integer]: IUserFunction read GetItem write SetItem; default;
    property ToString: String read GetToString;
    property Package: TUserPackage read FPackage;
  published
    { Published Declarations }
  end; { TUserFunctionList }

  TUserPackage = class(TCustomReasonerOwner)
  { Purpose: to provide a library or package of User Functions to allow easy
    integration with a reasoning engine at design time }
  private
    { Private Declarations }
    FEngine: TInferenceEngine;
    FFunctions: TUserFunctionList;
    FOnCreateFunctions: TNotifyEvent;
    procedure SetEngine( const Value: TInferenceEngine );
    function GetReset: Boolean;
    procedure SetReset(const Value: Boolean);
    procedure SetFunctions(const Value: TUserFunctionList);
  protected
    { Protected Declarations }
    procedure ReasonerNotification( Item: IReasoner; Action: TListNotification ); override;
    procedure FunctionNotification(AFunction: IUserFunction; Action: TListNotification); virtual;
    procedure AddFunctions( Reasoner: IReasoner ); virtual;
    procedure AddReasoners( AFunction: IUserFunction ); virtual;
    procedure RemoveReasoners( AFunction: IUserFunction ); virtual;
    procedure RemoveFunctions( Reasoner: IReasoner ); virtual;
    procedure CreateFunctions; virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DefineProperties(Filer: TFiler); override;
    procedure LoadFunctionsProperty(Reader: TReader);
    procedure StoreFunctionsProperty(Writer: TWriter);
  public
    { Public Declarations }
    procedure Assign( Source: TPersistent ); override;
    function IndexOf( AFunction: String ): Integer;
    procedure AfterConstruction; override;
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
    property Reasoners;
  published
    { Published Declarations }
    property Functions: TUserFunctionList read FFunctions write SetFunctions stored False;
    property Engine: TInferenceEngine read FEngine write SetEngine;
    property OnCreateFunctions: TNotifyEvent read FOnCreateFunctions write FOnCreateFunctions;
    property Reset: Boolean read GetReset write SetReset stored False;
  end; { TUserPackage }
  TUserPackageClass = class of TUserPackage;

  TFunctionCall = class(TPersistent, IFunCall, IUnknown)
  { Purpose:  To map a UserFunction to its arguments }
  private
    { Private Declarations }
    FReasoner: IReasoner;
    FArguments: TIList;
    FUserFunction: IUserFunction;
    FName: String;
    FRefCount: Integer;
    FTag: Integer;
    function GetArgumentCount: Integer;
    function GetReasoner: IReasoner;
  protected
    { Protected Declarations }
    function GetToString: String; virtual;
    function GetHashCode: Integer; virtual;
    function SimpleExecute( AContext: TIEContext ): IIEValue; virtual;
    property Arguments: TIList read FArguments;
  public
    { Public Declarations }
    procedure Assign( AValue: IUnknown ); reintroduce; overload;
    function Clone: IUnknown; virtual;
    constructor Create; overload; virtual;
    constructor Create( AReasoner: IReasoner; AName: TIEName ); overload;
    destructor Destroy; override;
    { IUnknown }
    property RefCount: Integer read FRefCount;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    { IFunCall }
    function AddArgument( Value: IIEValue ): Integer; overload;
    function AddArgument( Value: Double; Kind: TIEType ): Integer; overload;
    function AddArgument( Value: Integer; Kind: TIEType ): Integer; overload;
    function AddArgument( Value: String; Kind: TIEType ): Integer; overload;
    function AddArgument( Fact: IFact ): Integer; overload;
    function AddArgument( Value: IFunCall ): Integer; overload;
    function AddArgument( Value: IValueVector ): Integer; overload;
    function Equals( AValue: IUnknown ): Boolean; overload; virtual;
    function Equals( Value: IFunCall ): Boolean; overload;
    function GetTag: Integer;
    procedure SetTag(const Value: Integer);
    function GetArgument(Index: Integer): IIEValue;
    procedure SetArgument(Index: Integer; const Value: IIEValue);
    function GetName: TIEName;
    function GetUserFunction: IUserFunction;
    procedure SetUserFunction( const Value: IUserFunction );
    function Execute( AContext: TObject ): IIEValue; virtual;
    procedure Clear; virtual;
    procedure Delete(Index: Integer); virtual;
    property ArgumentCount: Integer read GetArgumentCount;
    property Argument[Index: Integer]: IIEValue read GetArgument write SetArgument;
    property UserFunction: IUserFunction read FUserFunction write SetUserFunction;

    property Name: String read FName;
    property HashCode: Integer read GetHashCode;
    property Reasoner: IReasoner read FReasoner;
    property ToString: String read GetToString;
  published
    { Published Declarations }
    property Tag: Integer read FTag write SetTag;
  end; { TFunctionCall }

  TInterpretedFunction = class(TIEUserFunction, IInterpretedFunction, IUserFunction, IUnknown)
  { Purpose:  Implements a DefFunction in CLIPS, in other words, a function
    defined by the CLIPS language.  Hence, it is interpreted }
  private
    { Private Declarations }
    FArguments: IValueVector;
    FStatements: IValueVector;
  protected
    { Protected Declarations }
    procedure Call( FunCall: IFunCall; Context: TIEContext;
              var Result: IIEValue ); overload; override;
    property Statements: IValueVector read FStatements;
    property Arguments: IValueVector read FArguments;
  public
    { Public Declarations }
    constructor Create; overload; override;
    destructor Destroy; override;
    procedure AfterConstruction; override;
    function GetMaxArgumentCount: Integer; override;
    function GetMinArgumentCount: Integer; override;
    // IInterpretedFunction
    procedure AddArgument( Name: String; AType: TIEType ); virtual;
    procedure AddStatement( FunCall: IFunCall ); overload; virtual;
    procedure AddStatement( Value: IIEValue ); overload; virtual;
    function GetArgument(Index: Integer): IIEValue;
    function GetArgumentCount: Integer;
    function GetStatement(Index: Integer): IIEValue;
    function GetStatementCount: Integer;
    function Execute( FunCall: IFunCall; AContext: TIEContext ): IIEValue; virtual;
    property Statement[Index: Integer]: IIEValue read GetStatement;
    property StatementCount: Integer read GetStatementCount;
    property Argument[Index: Integer]: IIEValue read GetArgument;
    property ArgumentCount: Integer read GetArgumentCount;
  published
    { Published Declarations }
  end; { TInterpretedFunction }

implementation

uses
    IEValue, IEClipsParser, IEHelperFunctions, TypInfo, IEFacts;

{ TCustomUserFunction }

procedure TCustomUserFunction.Assign(Source: TPersistent);
begin
     if Source is TCustomUserFunction then
     begin
          Comment := TCustomUserFunction(Source).Comment;
          FunctionName := TCustomUserFunction(Source).FunctionName;
          MaxArgumentCount := TCustomUserFunction(Source).MaxArgumentCount;
          MinArgumentCount := TCustomUserFunction(Source).MinArgumentCount;
          Exit;
     end;
     inherited Assign( Source );
end;

procedure TCustomUserFunction.Call(FunCall: IFunCall;
  Context: TIEContext; var Result: IIEValue);
begin
     result := nil;
     if Assigned( FOnCall ) then FOnCall( Self, FunCall, Context, Result );
end;

constructor TCustomUserFunction.Create(AOwner: TComponent; AName: TIEName);
{ convenience constructor to set the name }
begin
     Create( AOwner );
     FunctionName := AName;
end;

constructor TCustomUserFunction.Create(AOwner: TComponent);
begin
     inherited Create(AOwner);
     FInnerFunction := TIEInnerFunction.Create( Self );
     FInnerFunction.FunctionName := Name;
end;

destructor TCustomUserFunction.Destroy;
begin
     Engine := nil;
     inherited Destroy;
end;

function TCustomUserFunction.GetComment: String;
begin
     result := InnerFunction.Comment;
end;

function TCustomUserFunction.GetFunctionName: TIEName;
begin
     result := InnerFunction.FunctionName;
end;

function TCustomUserFunction.GetMaxArgumentCount: Integer;
begin
     result := InnerFunction.MaxArgumentCount;
end;

function TCustomUserFunction.GetMinArgumentCount: Integer;
begin
     result := InnerFunction.MinArgumentCount;
end;

procedure TCustomUserFunction.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
     inherited Notification( AComponent, Operation );
     if (Operation = opRemove) and
        (AComponent = Engine) then
        Engine := nil;
end;

procedure TCustomUserFunction.SetComment(const Value: String);
begin
     InnerFunction.Comment := Value;
end;

procedure TCustomUserFunction.SetEngine(const Value: TInferenceEngine);
begin
     if Value <> Engine then
     begin
          if csDesigning in ComponentState then
              FEngine := Value
          else
          begin
              if (FunctionName <> '') and (Engine <> nil) then
                 Engine.RemoveUserFunction( InnerFunction );
              FEngine := Value;
              if (FunctionName <> '') and (Engine <> nil) then
                 Engine.AddUserFunction( InnerFunction );
          end;
     end;
end;

procedure TCustomUserFunction.SetFunctionName(const NewName: TIEName);
begin
     InnerFunction.FunctionName := NewName;
end;

procedure TCustomUserFunction.SetMaxArgumentCount(const Value: Integer);
begin
     InnerFunction.MaxArgumentCount := Value;
end;

procedure TCustomUserFunction.SetMinArgumentCount(const Value: Integer);
begin
     InnerFunction.MinArgumentCount := Value;
end;

{ TUserPackage }

procedure TUserPackage.AddFunctions( Reasoner: IReasoner );
var
   i: Integer;
begin
     if (Reasoner <> nil) then
        for i := 0 to Functions.Count - 1 do
            Functions[i].AddReasoner( Reasoner );
end;

procedure TUserPackage.RemoveFunctions( Reasoner: IReasoner );
var
   i, j: Integer;
begin
     if Reasoner <> nil then
        for i := 0 to Functions.Count - 1 do
        begin
             j := Functions[i].IndexOfReasoner( Reasoner );
             if j > -1 then
                Functions[i].DeleteReasoner( j );
        end;
end;

constructor TUserPackage.Create(AOwner: TComponent);
var
   AReasoner: IReasoner;
begin
     inherited Create( AOwner );
     FFunctions := TUserFunctionList.Create(Self);
     if AOwner <> nil then
        if AOwner.GetInterface( IReasoner, AReasoner ) then
           Reasoners.Add( AReasoner );
end;

destructor TUserPackage.Destroy;
begin
     Reasoners.Clear;
     Functions.Clear;
     FEngine := nil;
     FFunctions.Free;
     FFunctions := nil;
     inherited Destroy;
end;

function TUserPackage.IndexOf(AFunction: String): Integer;
var
   i: Integer;
begin
     for i := 0 to Functions.Count - 1 do
         if Functions[i].FunctionName = AFunction then
         begin
              result := i;
              Exit;
         end;
     result := -1;
end;

procedure TUserPackage.SetEngine(const Value: TInferenceEngine);
begin
     if Value <> Engine then
     begin
          if csDesigning in ComponentState then
              FEngine := Value
          else
          begin
              if (Engine <> nil) and (Reasoners <> nil) then
                 Reasoners.Remove( Engine as IReasoner );
              FEngine := Value;
              if (Engine <> nil) and (Reasoners <> nil) then
                 Reasoners.Add( Engine as IReasoner );
          end;
     end;
end;

procedure TUserPackage.CreateFunctions;
begin
     if Assigned( FOnCreateFunctions ) then FOnCreateFunctions( Self );
end;

function TUserPackage.GetReset: Boolean;
begin
     result := False;
end;

procedure TUserPackage.SetReset(const Value: Boolean);
begin
     if Value then
     begin
          FFunctions.Clear;
          CreateFunctions;
     end;
end;

procedure TUserPackage.FunctionNotification(AFunction: IUserFunction;
  Action: TListNotification);
begin
     if (Reasoners = nil) or (Reasoners.Count = 0) then Exit;
     case Action of
       lnAdded:     AddReasoners( AFunction );
       lnExtracted,
       lnDeleted:   RemoveReasoners( AFunction );
     end;
end;

procedure TUserPackage.SetFunctions(const Value: TUserFunctionList);
begin
  FFunctions.Assign(Value);
end;

procedure TUserPackage.Assign(Source: TPersistent);
begin
     if Source is TUserPackage then
     begin
          FFunctions.Assign( TUserPackage(Source).Functions );
     end;
     inherited Assign( Source );
end;

procedure TUserPackage.Notification(AComponent: TComponent;
  Operation: TOperation);
var
   Reasoner: IReasoner;
begin
     inherited Notification( AComponent, Operation );
     if Operation = opRemove then
        if AComponent = Engine then
           Engine := nil
        else if (Reasoners <> nil) and
                AComponent.GetInterface( IReasoner, Reasoner ) then
             Reasoners.Remove( Reasoner );
end;

procedure TUserPackage.ReasonerNotification(Item: IReasoner;
  Action: TListNotification);
begin
     inherited;
     if Functions = nil then Exit;
     case Action of
       lnAdded:     AddFunctions( Item );
       lnExtracted,
       lnDeleted:
       begin
            RemoveFunctions( Item );
            if Item = (Engine as IReasoner) then
               FEngine := nil;
       end;
     end;
end;

procedure TUserPackage.AddReasoners(AFunction: IUserFunction);
var
   i: Integer;
begin
     for i := 0 to Reasoners.Count - 1 do
         AFunction.AddReasoner( Reasoners[i] );
end;

procedure TUserPackage.RemoveReasoners(AFunction: IUserFunction);
var
   i: Integer;
begin
     for i := 0 to Reasoners.Count - 1 do
         AFunction.AddReasoner( Reasoners[i] );
end;

procedure TUserPackage.DefineProperties(Filer: TFiler);
begin
     inherited; { allow base classes to define properties }
     Filer.DefineProperty('PackageFunctions', LoadFunctionsProperty, StoreFunctionsProperty,
                                              True);
end;

procedure TUserPackage.LoadFunctionsProperty(Reader: TReader);
var
   UF: TPersistent;
   Streamable: IDTStreamable;
begin
     Functions.Clear;
     Reader.ReadListBegin;
     while not Reader.EndOfList do
     begin
          // Read class name and create
          UF := FindClass(Reader.ReadString).Create;
          if UF.GetInterface( IDTStreamable, Streamable ) then
          begin
               Streamable.ReadData( Reader );
               Functions.Add( Streamable as IUserFunction );
          end;
     end;
     Reader.ReadListEnd;
end;

procedure TUserPackage.StoreFunctionsProperty(Writer: TWriter);
var
   i: Integer;
   Streamable: IDTStreamable;
begin
     Writer.WriteListBegin;
     try
        // Stream out any functions that implement the IDTStreamable interface
        for i := 0 to Functions.Count - 1 do
            if Functions[i].QueryInterface( IDTStreamable, Streamable ) <> E_NoInterface then
            begin
                 Writer.WriteString( Streamable.GetClassName );
                 Streamable.WriteData( Writer );
            end;
     finally
        Writer.WriteListEnd;
     end;
end;

procedure TUserPackage.AfterConstruction;
begin
     inherited AfterConstruction;
     CreateFunctions;
end;

{ TFunctionCall }

function TFunctionCall.AddArgument(Value: String; Kind: TIEType): Integer;
var
   v: IIEValue;
begin
     v := TIEValue.Create;
     case Kind of
          dtAtom:                         v.AsAtom := Value;
          dtString:                       v.AsString := Value;
          dtVariable:                     v.AsVariable := Value;
          dtMultiVariable:                v.AsMultiVariable := Value;
          dtSlot:                         v.AsSlot := Value;
          dtMultiSlot:                    v.AsMultiSlot := Value;
     else
         raise EIEException.CreateFmt( SNotCorrectType,
                                      ['String type',
                                      GetEnumName(TypeInfo(TIEType),Ord(Kind))]);
     end;
     result := FArguments.Add( v );
end;

function TFunctionCall.AddArgument(Value: Integer; Kind: TIEType): Integer;
var
   v: IIEValue;
begin
     v := TIEValue.Create;
     case Kind of
          dtInteger:                      v.AsInteger := Value;
          dtFloat:                        v.AsFloat := Value;
     else
         raise EIEException.CreateFmt( SNotCorrectType,
                                      ['Integer type',
                                      GetEnumName(TypeInfo(TIEType),Ord(Kind))]);
     end;
     result := FArguments.Add( v );
end;

function TFunctionCall.AddArgument(Value: Double; Kind: TIEType): Integer;
var
   v: IIEValue;
begin
     v := TIEValue.Create;
     case Kind of
          dtInteger:                      v.AsInteger := trunc(Value);
          dtFloat:                        v.AsFloat := Value;
     else
         raise EIEException.CreateFmt( SNotCorrectType,
                                      ['Float type',
                                      GetEnumName(TypeInfo(TIEType),Ord(Kind))]);
     end;
     result := FArguments.Add( v );
end;

function TFunctionCall.AddArgument(Value: IFunCall): Integer;
var
   v: IIEValue;
begin
     v := TIEValue.Create;
     v.AsFunCall := Value;
     result := FArguments.Add( v );
end;

function TFunctionCall.AddArgument(Fact: IFact): Integer;
var
   v: IIEValue;
begin
     v := TIEValue.Create;
     v.AsFact := Fact;
     result := FArguments.Add( v );
end;

function TFunctionCall.AddArgument(Value: IIEValue): Integer;
begin
     result := FArguments.Add( Value );
end;

function TFunctionCall.AddArgument(Value: IValueVector): Integer;
var
   v: IIEValue;
begin
     v := TIEValue.Create;
     v.AsList := Value;
     result := FArguments.Add( v );
end;

procedure TFunctionCall.Assign(AValue: IUnknown);
var
   Value: IIEValue;
   i: Integer;
begin
     with AValue as IFunCall do
     begin
          Self.FName := Name;
          Self.FTag := Tag;
          Self.Arguments.Clear;
          for i := 0 to ArgumentCount - 1 do
          begin
               Value := (Argument[i] as IIEValue).Clone as IIEValue;
               Self.FArguments.Add( Value as IIEValue );
          end;
     end;
end;

function TFunctionCall.Clone: IUnknown;
var
   AResult: TFunctionCall;
begin
     AResult := TFunctionCall.Create( Reasoner, Name );
     AResult.Assign( Self );
     result := AResult;
end;

constructor TFunctionCall.Create(AReasoner: IReasoner; AName: TIEName);
begin
     Create;
     FReasoner := AReasoner;
     FName := AName;
end;

destructor TFunctionCall.Destroy;
begin
     FReasoner := nil;
     FArguments.Free;
     FArguments := nil;
     inherited Destroy;
end;

function TFunctionCall.Equals(Value: IFunCall): Boolean;
var
   i: Integer;
begin
     result := (Name = Value.Name) and
               (ArgumentCount = Value.ArgumentCount);
     if result then
     begin
          for i := 0 to ArgumentCount - 1 do
              if not Argument[i].Equals( Value.Argument[i] ) then
              begin
                   result := False;
                   Exit;
              end;
     end;
end;

function TFunctionCall.Execute(AContext: TObject): IIEValue;
begin
     // Make sure we have a userfunction
     if (UserFunction = nil) and (Reasoner <> nil) then
        if not Reasoner.ContainsUserFunction(Name) then
           raise EIEException.CreateFmt( SNoSuchFunction, [Name] )
        else
            UserFunction := Reasoner.UserFunction[Name];
     UserFunction.Call( Self, AContext as TIEContext, Result );
end;

function TFunctionCall.GetArgument(Index: Integer): IIEValue;
begin
     result := FArguments[Index] as IIEValue;
end;

function TFunctionCall.GetArgumentCount: Integer;
begin
     result := Arguments.Count;
end;

function TFunctionCall.GetName: TIEName;
begin
     result := FName;
end;

function TFunctionCall.GetToString: String;
begin
     if Reasoner <> nil then
        result := Reasoner.PrettyPrint( Self )
     else
         result := IEClipsParser.PrettyPrint( Self );
end;

function TFunctionCall.GetUserFunction: IUserFunction;
begin
     result := FUserFunction;
end;

procedure TFunctionCall.SetArgument(Index: Integer; const Value: IIEValue);
{ Set the Argument, an exception is thrown if index is out of bounds }
begin
     FArguments[Index] := Value;
end;

procedure TFunctionCall.SetUserFunction(const Value: IUserFunction);
begin
     if FUserFunction <> Value then
     begin
          FUserFunction := Value;
          if UserFunction <> nil then
             FName := UserFunction.FunctionName
          else
              FName := '';
     end;
end;

function TFunctionCall.SimpleExecute(AContext: TIEContext): IIEValue;
begin
     if UserFunction = nil then
        raise EIEException.CreateFmt( SNoSuchFunction, [Name] );
     try
        UserFunction.Call( Self, AContext, Result );
     except
        On E: Exception do
           raise EIEException.CreateFmt( SFunctionError, [Name,E.Message] );
     end;
end;

function TFunctionCall._AddRef: Integer;
begin
  Result := InterlockedIncrement(FRefCount);
end;

function TFunctionCall._Release: Integer;
begin
  result := InterlockedDecrement(FRefCount);
  if result = 0 then
  begin
       Free;
       Exit;
  end;
end;

function TFunctionCall.QueryInterface(const IID: TGUID; out Obj): HResult;
const
  E_NOINTERFACE = HResult($80004002);
begin
  if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

function TFunctionCall.GetHashCode: Integer;
var
   i: Integer;
begin
     result := CalculateHash( Name );
     for i := 0 to ArgumentCount - 1 do
         result := result + Argument[i].HashCode;
end;

function TFunctionCall.Equals(AValue: IUnknown): Boolean;
var
   FC: IFunCall;
begin
     result := AValue.QueryInterface( IFunCall, FC ) <> E_NOINTERFACE;
     if result then
        result := Equals( FC );
end;

function TFunctionCall.GetTag: Integer;
begin
     result := FTag;
end;

procedure TFunctionCall.SetTag(const Value: Integer);
begin
     FTag := Value;
end;

function TFunctionCall.GetReasoner: IReasoner;
begin
     result := FReasoner;
end;

procedure TFunctionCall.Clear;
// Clears the function call of all arguments
begin
     FArguments.Clear;
end;

procedure TFunctionCall.Delete(Index: Integer);
// Deletes an argument
begin
     FArguments.Delete(Index);
end;

constructor TFunctionCall.Create;
begin
     inherited Create;
     FArguments := TIList.Create;
end;

{ TInterpretedFunction }

procedure TInterpretedFunction.AddStatement(FunCall: IFunCall);
begin
     Statement[FStatements.Add( TIEValue.Create )].AsFunCall := FunCall;
end;

procedure TInterpretedFunction.AddArgument(Name: String; AType: TIEType);
{ Type better be just dtAtom (LOCAL), dtVariable, or dtMultiVariable }
begin
     if not (AType in [dtAtom, dtVariable, dtMultiVariable]) then
        raise EIEException.CreateFmt( SInvalidArgType, [GetEnumName(TypeInfo(TIEType),Ord(AType))] );
     with FArguments[FArguments.Add( TIEValue.Create )] do
          case AType of
               dtAtom:           AsAtom := Name;
               dtVariable:       AsVariable := Name;
               dtMultiVariable:  AsMultiVariable := Name;
          end;
end;

procedure TInterpretedFunction.AddStatement(Value: IIEValue);
begin
     FStatements.Add( Value );
end;

procedure TInterpretedFunction.Call(FunCall: IFunCall; Context: TIEContext;
  var Result: IIEValue);
{ Execute the interpreted functions.  For every FunCall in Statement list,
  expand it and execute it }
var
   i, j: Integer;
   Multi: IValueVector;
   Value: IIEValue;
begin
     inherited Call( FunCall, Context, Result );
     // Make sure we have enough arguments
     if FunCall.ArgumentCount < Arguments.ValueCount then
        raise EIEException.CreateFmt( STooFewArgs, [FunctionName] );

     // Ok, we need our own local context for the parameters and other variables
     Context := Context.Push;
     try
        // Initialize the variables
        for i := 0 to Arguments.ValueCount - 1 do
        begin
          case Argument[i].DType of
               dtAtom:           // Local variable, initialize
                  Context.Variable[Arguments[i].AsString] := FCNil;
               dtVariable:       // Parameter, bind to an argument
                  Context.AddLocalBinding(Arguments[i].AsString, FunCall.Argument[i].Resolve(Context) );
               dtMultiVariable:  // Parameter, bind to the last few arguments
               begin
                    Multi := TValueVector.Create;
                    for j := i to FunCall.ArgumentCount - 1 do
                        Multi.Add( FunCall.Argument[j].Resolve(Context) );
                    Value := TIEValue.Create;
                    Value.AsList := Multi;
                    Context.AddLocalBinding(Arguments[i].AsString, Value);
               end;
          end;
        end;

        // Execute the function by doing all the actions
        for i := 0 to Statements.ValueCount - 1 do
        begin
          result := Statements[i].Resolve(Context);
          // returning the value? then we are done
          if Context.Returning then
          begin
               result := Context.ReturnValue;
               Context.ClearReturnValue;
               Break;
          end;
        end;
     finally
        Context.Pop;
        Context.Free;
     end;
     result := result.Resolve(Context);
end;

constructor TInterpretedFunction.Create;
begin
     inherited Create;
     FStatements := TValueVector.Create;
     FArguments := TValueVector.Create;
end;

destructor TInterpretedFunction.Destroy;
begin
     FStatements := nil;
     FArguments := nil;
     inherited Destroy;
end;

function TInterpretedFunction.Execute(FunCall: IFunCall; AContext: TIEContext): IIEValue;
// execute the interpreted function
begin
     Call( FunCall, AContext, result );
end;

function TInterpretedFunction.GetStatement(Index: Integer): IIEValue;
begin
     result := FStatements[Index];
end;

function TInterpretedFunction.GetStatementCount: Integer;
begin
     result := FStatements.ValueCount;
end;

function TInterpretedFunction.GetArgument(Index: Integer): IIEValue;
begin
     result := Arguments[Index];
end;

function TInterpretedFunction.GetArgumentCount: Integer;
begin
     result := FArguments.ValueCount;
end;

function TInterpretedFunction.GetMaxArgumentCount: Integer;
begin
     if (Arguments.ValueCount > 0) and
        (Argument[Arguments.ValueCount-1].DType = dtMultiVariable)  then
        result := MaxInt
     else
         result := Arguments.ValueCount;
end;

function TInterpretedFunction.GetMinArgumentCount: Integer;
begin
     result := Arguments.ValueCount;
end;

procedure TInterpretedFunction.AfterConstruction;
begin
     inherited AfterConstruction;
     // if FStatements = nil then create was not called (happens) when the object
     // is created by FindClass(ClassName).Create.  The only place this is done
     // right now is when streaming so the read data will take care of fixing
     // the other properties
     if Statements = nil then
     begin
          FStatements := TValueVector.Create;
          FArguments := TValueVector.Create;
     end;
end;

{ TUserFunctionList }

function TUserFunctionList.Add(AFunction: IUserFunction): Integer;
begin
     result := IndexOf( AFunction );
     if result = -1 then
        result := inherited Add( AFunction );
end;

constructor TUserFunctionList.Create(Package: TUserPackage);
begin
{$IFDEF DEMO}
     inherited Create( @FHeap, High(FHeap)+1 );
{$ELSE}
     inherited Create;
{$ENDIF}
     FPackage := Package;
end;

function TUserFunctionList.GetItem(Index: Integer): IUserFunction;
begin
     result := inherited Items[Index] as IUserFunction;
end;

function TUserFunctionList.GetToString: String;
var
   i: Integer;
begin
     result := '';
     for i := 0 to Count - 1 do
         result := result + Items[i].FunctionName + NewLineString;
end;

function TUserFunctionList.IndexOf(AFunction: IUserFunction): Integer;
begin
     result := inherited IndexOf( AFunction );
end;

procedure TUserFunctionList.Insert(Index: Integer; AFunction: IUserFunction);
begin
     inherited Insert( Index, AFunction );
end;

procedure TUserFunctionList.Notify(Item: IUnknown; Action: TListNotification);
begin
     if Package <> nil then
        Package.FunctionNotification( Item as IUserFunction, Action );
     inherited Notify( Item, Action );
end;

function TUserFunctionList.Remove(AFunction: IUserFunction): Integer;
begin
     result := inherited Remove( AFunction );
end;

procedure TUserFunctionList.SetItem(Index: Integer;
  const Value: IUserFunction);
begin
     inherited Items[Index] := Value;
end;

{ TBaseReasonerList }

function TBaseReasonerList.Add(Reasoner: IReasoner): Integer;
begin
     result := IndexOf( Reasoner );
     if result = -1 then
        result := inherited Add( Reasoner );
end;

constructor TBaseReasonerList.Create(AOwner: TIEUserFunction);
begin
     inherited Create;
     FOwner := AOwner;
end;

function TBaseReasonerList.GetItem(Index: Integer): IReasoner;
begin
     result := inherited Items[Index] as IReasoner;
end;

function TBaseReasonerList.IndexOf(Reasoner: IReasoner): Integer;
begin
     result := inherited IndexOf(Reasoner);
end;

procedure TBaseReasonerList.Insert(Index: Integer; Reasoner: IReasoner);
begin
     inherited Insert( Index, Reasoner );
end;

procedure TBaseReasonerList.Notify(Item: IUnknown; Action: TListNotification);
begin
     inherited;
     Owner.ReasonerNotification( Item as IReasoner, Action );
end;

function TBaseReasonerList.Remove(Reasoner: IReasoner): Integer;
begin
     result := inherited Remove(Reasoner);
end;

procedure TBaseReasonerList.SetItem(Index: Integer; const Value: IReasoner);
begin
     inherited Items[Index] := Value;
end;

{ TCustomReasonerOwner }

constructor TCustomReasonerOwner.Create(AOwner: TComponent);
begin
     inherited Create(AOwner);
     FReasoners := TReasonerList.Create(Self);
end;

destructor TCustomReasonerOwner.Destroy;
begin
     Reasoners.Clear;
     FReasoners.Free;
     FReasoners := nil;
     inherited Destroy;
end;

procedure TCustomReasonerOwner.ReasonerNotification(Item: IReasoner;
  Action: TListNotification);
begin

end;

procedure TCustomReasonerOwner.SetReasoners(const Value: TReasonerList);
begin
     FReasoners.Assign(Value);
end;

{ TReasonerList }

constructor TReasonerList.Create(AOwner: TCustomReasonerOwner);
begin
     inherited Create(nil);
     FOwner := AOwner;
end;

procedure TReasonerList.Notify(Item: IUnknown; Action: TListNotification);
begin
     //  inherited;  Don't do inherited
     Owner.ReasonerNotification( Item as IReasoner, Action );
end;

{ TIEUserFunction }

function TIEUserFunction.AddReasoner(AReasoner: IReasoner): Integer;
begin
     result := FReasoners.Add(AReasoner);
end;

procedure TIEUserFunction.AfterConstruction;
begin
     inherited AfterConstruction;
     // if reasoners = nil then create was not called (happens) when the object
     // is created by FindClass(ClassName).Create.  The only place this is done
     // right now is when streaming so the read data will take care of fixing
     // the other properties
     if Reasoners = nil then
        FReasoners := TBaseReasonerList.Create( Self );
     // Release the constructor's implicit refcount
     InterlockedDecrement(FRefCount);
end;

procedure TIEUserFunction.BeforeDestruction;
begin
     inherited BeforeDestruction;
     if RefCount <> 0 then raise EIEException.Create(SInvalidPtr);
end;

procedure TIEUserFunction.Call(FunCall: IFunCall; Context: TObject;
  var Result: IIEValue);
begin
     Call( FunCall, Context as TIEContext, Result );
end;

procedure TIEUserFunction.Call(FunCall: IFunCall; Context: TIEContext;
  var Result: IIEValue);
begin
     // Make sure we have enough arguments
     if FunCall.ArgumentCount < MinArgumentCount then
        raise EIEException.CreateFmt( STooFewArgs, [FunctionName] )
     else if FunCall.ArgumentCount > MaxArgumentCount then
        raise EIEException.CreateFmt( STooManyArgs, [FunctionName] );
     result := nil;
end;

procedure TIEUserFunction.ClearReasoners;
begin
     while Reasoners.Count > 0 do
           Reasoners.Delete( 0 );
end;

constructor TIEUserFunction.Create;
begin
     inherited Create;
     FReasoners := TBaseReasonerList.Create( Self );
     FMaxArgumentCount := MaxInt;
end;

procedure TIEUserFunction.DeleteReasoner(Index: Integer);
begin
     FReasoners.Delete(Index);
end;

constructor TIEUserFunction.Create(AName: TIEName);
begin
     Create;
     FunctionName := AName;
end;

procedure TIEUserFunction.DeleteReasoner(AReasoner: IReasoner);
begin
     FReasoners.Remove( AReasoner );
end;

destructor TIEUserFunction.Destroy;
begin
     FReasoners.Free;
     inherited;
end;

function TIEUserFunction.GetComment: String;
begin
     result := FComment;
end;

function TIEUserFunction.GetFunctionName: TIEName;
begin
     result := FFunctionName;
end;

function TIEUserFunction.GetMaxArgumentCount: Integer;
begin
     result := FMaxArgumentCount;
end;

function TIEUserFunction.GetMinArgumentCount: Integer;
begin
     result := FMinArgumentCount;
end;

function TIEUserFunction.GetReasoner(Index: Integer): IReasoner;
begin
     result := Reasoners[Index];
end;

function TIEUserFunction.GetReasonerCount: Integer;
begin
     result := Reasoners.Count;
end;

function TIEUserFunction.GetToString(const Reasoner: IReasoner): String;
begin
     result := Reasoner.PrettyPrint( Self );
end;

function TIEUserFunction.IndexOfReasoner(AReasoner: IReasoner): Integer;
begin
     result := FReasoners.IndexOf( AReasoner );
end;

class function TIEUserFunction.NewInstance: TObject;
begin
     // Set an implicit refcount so that refcounting
     // during construction won't destroy the object.
     Result := inherited NewInstance;
     TIEUserFunction(Result).FRefCount := 1;
end;
 
function TIEUserFunction.QueryInterface(const IID: TGUID;
  out Obj): HResult;
const
  E_NOINTERFACE = HResult($80004002);
begin
     if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

procedure TIEUserFunction.ReasonerNotification(Item: IReasoner;
  Action: TListNotification);
begin
     case Action of
       lnAdded:    Item.AddUserFunction( Self );
       lnExtracted,
       lnDeleted:  Item.RemoveUserFunction(Self);
     end;
end;

procedure TIEUserFunction.SetComment(const Value: String);
begin
     FComment := Value;
end;

procedure TIEUserFunction.SetFunctionName(const Value: TIEName);
var
   i: Integer;
begin
     if FFunctionName <> Value then
     begin
          for i := 0 to Reasoners.Count - 1 do
          begin
               Reasoners[i].RemoveUserFunction( Self );
          end;
          FFunctionName := Value;
          for i := 0 to Reasoners.Count - 1 do
          begin
               Reasoners[i].AddUserFunction( Self );
          end;
     end;
end;

procedure TIEUserFunction.SetMaxArgumentCount(const Value: Integer);
begin
     if FMaxArgumentCount <> Value then
     begin
          FMaxArgumentCount := Value;
          if Value < FMinArgumentCount then
             FMinArgumentCount := Value;
     end;
end;

procedure TIEUserFunction.SetMinArgumentCount(const Value: Integer);
begin
     if FMinArgumentCount <> Value then
     begin
          FMinArgumentCount := Value;
          if Value > FMaxArgumentCount then
             FMaxArgumentCount := Value;
     end;
end;

procedure TIEUserFunction.SetReasoners(const Value: TBaseReasonerList);
begin
     FReasoners.Assign( Value );
end;

function TIEUserFunction._AddRef: Integer;
begin
     Result := InterlockedIncrement(FRefCount);
end;

function TIEUserFunction._Release: Integer;
begin
     Result := InterlockedDecrement(FRefCount);
     if Result = 0 then
        Destroy;
end;

procedure TIEUserFunction.Assign(Source: TPersistent);
begin
     if Source is TIEUserFunction then
     begin
          FMinArgumentCount := TIEUserFunction(Source).FMinArgumentCount;
          FMaxArgumentCount := TIEUserFunction(Source).FMaxArgumentCount;
          FComment := TIEUserFunction(Source).FComment;
          FunctionName := TIEUserFunction(Source).FunctionName;
          Exit;
     end;
     inherited Assign( Source );
end;

procedure TIEUserFunction.ReadData(Reader: TReader);
// Ugly hack because nothing else works :(
begin
     FunctionName := Reader.ReadString;
     Comment := Reader.ReadString;
     FMaxArgumentCount := Reader.ReadInteger;
     FMinArgumentCount := Reader.ReadInteger;
end;

procedure TIEUserFunction.WriteData(Writer: TWriter);
// Ugly hack because nothing else works :(
begin
     Writer.WriteString( FunctionName );
     Writer.WriteString( Comment );
     Writer.WriteInteger( MaxArgumentCount );
     Writer.WriteInteger( MinArgumentCount );
end;

function TIEUserFunction.GetClassName: String;
begin
     result := ClassName;
end;

{ TIEInnerFunction }

procedure TIEInnerFunction.Call(FunCall: IFunCall; Context: TIEContext;
  var Result: IIEValue);
begin
     inherited;
     if Owner = nil then
        raise EIEException.CreateFmt( SNoSuchFunction, [FunctionName] );
     if Owner is TCustomUserFunction then
        TCustomUserFunction(Owner).Call( FunCall, Context, Result )
     else if Assigned(FOnCall) then FOnCall( Self, FunCall, Context, Result );
end;

constructor TIEInnerFunction.Create(AOwner: TObject);
begin
     inherited Create;
     FOwner := AOwner;
     FNexus := TIEFunctionNexus.Create( Self );
     // if owner is a TComponent, we are going to tell it to notify us when
     // it is destroyed.  If it isn't, it is up to the owner to clear
     // OnCall event before destroying itself
     if AOwner is TComponent then
        TComponent(Owner).FreeNotification( FNexus );
end;

destructor TIEInnerFunction.Destroy;
begin
     FNexus.Free;
     FOwner := nil;
     inherited Destroy;
end;

procedure TIEInnerFunction.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
     if (Operation = opRemove) and
        (AComponent = Owner) then
        FOwner := nil;
end;

{ TIEFunctionNexus }

constructor TIEFunctionNexus.Create(InnerFunction: TIEInnerFunction);
begin
     inherited Create(nil);
     FInnerFunction := InnerFunction;
end;

procedure TIEFunctionNexus.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
     inherited Notification( AComponent, Operation );
     InnerFunction.Notification( AComponent, Operation );
end;

initialization
{
  Component writers call RegisterClass to register custom classes so that the
  class type can be obtained from the class name by calling the GetClass
  function. If a class is not registered, GetClass returns nil when passed the
  class name.  This is important for read/write from/to streams.
}
  RegisterClass( TCustomUserFunction );
  RegisterClass( TUserFunction );
  RegisterClass( TIEUserFunction );

end.
